from re import L
from .lambda_layers import call_lambda_layer
from .reader import get_graph_as_dict
import tensorflow as tf
from typing import Dict


def replace_layers_by_counterparts_in_model(
    model: tf.keras.Model,
    replace_layers_by_provided_counter_parts: Dict[str, tf.keras.layers.Layer],
    verbose: bool = False,
) -> tf.keras.Model:
    """
    replaces layers basd on types

    Args:
        model: model to edit
        replace_layers_by_provided_counter_parts: Dict (name, new layer)
        verbose: verbatim level
    """
    modified_layers = 0
    network_dict, is_transformer = get_graph_as_dict(model=model)
    model_outputs = []
    model_layers = model.layers[1:]
    layer_cpt_shift = 1
    model_outputnames = model.output_names
    if is_transformer:
        model_layers = model.layers
        model_outputnames = [model.layers[-1].name]
        layer_cpt_shift = 0

    for layer_cpt, layer in enumerate(model_layers):
        if verbose:
            print(
                f"\r[\033[96mReplaceLayersByCounterparts\033[0m] ["
                + "\033[91m=\033[0m" * int(15 * layer_cpt / len(model_layers))
                + " " * int(15 - int(15 * layer_cpt / len(model_layers)))
                + "]",
                end="",
            )
        if isinstance(layer, tf.keras.layers.InputLayer):
            layer_input = network_dict["input"][layer.name]
        else:
            input_layers = []
            for in_node in layer.inbound_nodes:
                if isinstance(in_node.inbound_layers, (list, tuple)):
                    for inl in in_node.inbound_layers:
                        input_layers.append(inl.name)
                else:
                    input_layers.append(in_node.inbound_layers.name)
            layer_input = [
                network_dict["new_output"][layer_aux] for layer_aux in input_layers
            ]
        if layer.name in network_dict["new_output"]:
            continue
        if len(layer_input) == 1:
            layer_input = layer_input[0]

        if isinstance(layer, tf.keras.layers.InputLayer):
            x = layer_input
        elif len(network_dict["input"][layer.name]) == 1:
            if layer.name in replace_layers_by_provided_counter_parts:
                modified_layers += 1
                x = replace_layers_by_provided_counter_parts[layer.name](layer_input)
            else:
                copied_layer = type(layer).from_config(layer.get_config())
                x = call_lambda_layer(
                    layer_input=layer_input,
                    model=model,
                    layer=copied_layer,
                    layer_cpt=layer_cpt + layer_cpt_shift,
                )
        else:
            if layer.name in replace_layers_by_provided_counter_parts:
                modified_layers += 1
                x = replace_layers_by_provided_counter_parts[layer.name](layer_input)
            elif isinstance(layer, tf.keras.layers.Multiply) or isinstance(
                layer, tf.keras.layers.Concatenate
            ):
                copied_layer = type(layer).from_config(layer.get_config())
                x = layer(
                    [
                        network_dict["new_output"][layer_aux]
                        for layer_aux in network_dict["input"][layer.name]
                    ]
                )
            else:
                copied_layer = type(layer).from_config(layer.get_config())
                if (
                    isinstance(layer, tf.keras.layers.Lambda)
                    or ("lambda" in type(layer).__name__.lower())
                    and "tf.reshape" in layer.name.lower()
                ):
                    x = copied_layer(
                        layer_input[0],
                        shape=(
                            tf.shape(layer_input[0])[0],
                            1,
                            tf.shape(layer_input[0])[1],
                        ),
                    )
                elif (
                    isinstance(layer, tf.keras.layers.Lambda)
                    or ("lambda" in type(layer).__name__.lower())
                    and "tf.broadcast_to" in layer.name.lower()
                ):
                    x = copied_layer(
                        layer_input[2],
                        shape=(
                            layer_input[0],
                            layer_input[3],
                            layer_input[1],
                        ),
                    )
                elif isinstance(layer, tf.keras.layers.MultiHeadAttention):
                    i_name, _, _, extra_args = model.get_config()["layers"][
                        layer_cpt + layer_cpt_shift
                    ]["inbound_nodes"][0][0]
                    new_layer_inputs = {}
                    for i in layer_input:
                        if extra_args["value"][0] in i.name:
                            new_layer_inputs["value"] = i
                        elif extra_args["attention_mask"][0] in i.name:
                            new_layer_inputs["attention_mask"] = i
                    x = copied_layer(
                        query=new_layer_inputs["value"],
                        value=new_layer_inputs["value"],
                        attention_mask=new_layer_inputs["attention_mask"],
                    )
                elif "tf.__operators__.add" in layer.name:
                    x = copied_layer(layer_input[0], layer_input[1])
                else:
                    x = copied_layer(layer_input)
        network_dict["new_output"].update({layer.name: x})
        if layer.name in model_outputnames:
            model_outputs.append(x)
    new_model_inputs = model.inputs
    if model.inputs is None:
        new_model_inputs = network_dict["new_output"][
            network_dict["input"][model.layers[0].name][0]
        ]
    if isinstance(model.inputs, (tuple, list)):
        new_model_inputs = model.inputs
    elif is_transformer:
        new_layer_inputs = network_dict["input"][model.layers[0].name][0]
    new_model = tf.keras.Model(inputs=new_model_inputs, outputs=model_outputs[-1])
    new_model._name = model.name
    unmodified_layers = len(model.layers) - modified_layers

    if verbose:
        print(
            f"\r[\033[96mReplaceLayersByCounterparts\033[0m] ["
            + "\033[96m=\033[0m" * 15
            + "]"
        )
        print("\n\r+" + "-" * 39 + "+")
        print(f'| {f"{model.name}".center(37)} |')
        print("\r+" + "-" * 39 + "+")
        print(f'| modified layers             | {f"{modified_layers}":<7} |')
        print(f'| unmodified layers           | {f"{unmodified_layers}":<7} |')
        print("\r+" + "-" * 39 + "+")
    return new_model



if __name__ == "__main__":
    import os

    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
from .utils.replace_type_editor import replace_layers_by_types_in_model
from .utils.remove_type_editor import remove_layers_by_types_in_model
from .utils.replace_by_counter import replace_layers_by_counterparts_in_model
from .utils.remove_by_names import remove_layers_by_names_or_indices_in_model
import tensorflow as tf
from typing import Optional, List, Tuple, Type, Dict, Union, Any


def edit_model(
    model: tf.keras.Model,
    remove_by_name: Optional[List[Union[str, int]]] = None,
    replace_layers_by_provided_counter_parts: Optional[
        Dict[str, tf.keras.layers.Layer]
    ] = None,
    replace_layers_by_types: Optional[List[Tuple[Type, Type]]] = None,
    remove_layer_by_types: Optional[List[Type]] = None,
    extra_args: Dict[str, Any] = None,
    verbose: bool = False,
) -> tf.keras.Model:
    """
    The purpose of this project is to provide an easy way to replace or remove layes in a tensorflow model.
    Note: this method will not apply pruning or tensor decomposition.


    Examples:

    >>> model = tf.keras.applications.MobileNetV2(weights=None)
    >>> new_model = edit_model(
    ...     model=model,
    ...     remove_by_name=[2, -3, "block_16_depthwise_BN"],
    ...     verbose=True,
    ... )
    [RemoveLayersByName] [===============]
    +---------------------------------------+
    |          mobilenetv2_1.00_224         |
    +---------------------------------------+
    | removed layers              | 3       |
    | kept layers                 | 153     |
    +---------------------------------------+


    >>> model = tf.keras.applications.MobileNetV2(weights=None)
    >>> new_model = edit_model(
    ...     model=model,
    ...     replace_layers_by_types=[
    ...         [tf.keras.layers.BatchNormalization, tf.keras.layers.LayerNormalization]
    ...     ],
    ...     verbose=True,
    ... )
    [ReplaceLayersByType] [===============]
    +---------------------------------------+
    |          mobilenetv2_1.00_224         |
    +---------------------------------------+
    | modified layers             | 52      |
    | unmodified layers           | 104     |
    +---------------------------------------+

    >>> model = tf.keras.applications.MobileNetV2(weights=None)
    >>> new_model = edit_model(
    ...     model=model,
    ...     remove_layer_by_types=[tf.keras.layers.BatchNormalization],
    ...     verbose=True,
    ... )
    [RemoveLayersByType] [===============]
    +---------------------------------------+
    |          mobilenetv2_1.00_224         |
    +---------------------------------------+
    | removed layers              | 52      |
    | kept layers                 | 104     |
    +---------------------------------------+

    >>> model = tf.keras.applications.MobileNetV2(weights=None)
    >>> new_model = edit_model(
    ...     model=model,
    ...     replace_layers_by_provided_counter_parts={
    ...         "Conv_1": tf.keras.layers.Conv2D(
    ...             filters=1280, kernel_size=5, padding="same"
    ...         )
    ...     },
    ...     verbose=True,
    ... )
    [ReplaceLayersByCounterparts] [===============]
    +---------------------------------------+
    |          mobilenetv2_1.00_224         |
    +---------------------------------------+
    | modified layers             | 1       |
    | unmodified layers           | 155     |
    +---------------------------------------+

    Args:
        model: model to edit
        remove_by_name: list of names (or indices) of layers to remove
        replace_layers_by_provided_counter_parts: Dict of pairs (layer name, new layer) and we replace
            the layer with correct name by the provided new layer
        replace_layers_by_types: list of pairs (old layer type, new layer type)
        remove_layer_by_types: list of types of layers to remove
        extra_args: extra arguments for new layers in replace by types.
        verbose: verbatim level
    """
    if replace_layers_by_types is not None:
        return replace_layers_by_types_in_model(
            model=model,
            replace_layers=replace_layers_by_types,
            extra_args=extra_args,
            verbose=verbose,
        )
    elif remove_layer_by_types is not None:
        return remove_layers_by_types_in_model(
            model=model,
            remove_layer=remove_layer_by_types,
            verbose=verbose,
        )
    elif replace_layers_by_provided_counter_parts is not None:
        return replace_layers_by_counterparts_in_model(
            model=model,
            replace_layers_by_provided_counter_parts=replace_layers_by_provided_counter_parts,
            verbose=verbose,
        )
    elif remove_by_name is not None:
        return remove_layers_by_names_or_indices_in_model(
            model=model, remove_layer=remove_by_name, verbose=verbose
        )
    else:
        print("\r[\033[91mwarning\033[0m] called tfsurgeon without modifying the model")
        return model


if __name__ == "__main__":
    model = tf.keras.applications.MobileNetV2(weights=None)
    new_model = edit_model(
        model=model,
        replace_layers_by_types=[
            [tf.keras.layers.BatchNormalization, tf.keras.layers.LayerNormalization]
        ],
        verbose=True,
    )
    new_model = edit_model(
        model=model,
        remove_layer_by_types=[tf.keras.layers.BatchNormalization],
        verbose=True,
    )
    new_model = edit_model(
        model=model,
        replace_layers_by_provided_counter_parts={
            "Conv_1": tf.keras.layers.Conv2D(
                filters=1280, kernel_size=5, padding="same"
            )
        },
        verbose=True,
    )
    new_model = edit_model(
        model=model,
        remove_by_name=[2, -3, "block_16_depthwise_BN"],
        verbose=True,
    )
    new_model = edit_model(
        model=model,
        verbose=True,
    )



from typing import Any, Dict
import tensorflow as tf
import numpy as np


def get_lambda_args(cpt: int, model: tf.keras.Model) -> Any:
    """
    retrieve the lambda arguments from model config as
    they can't be accessed from the layer config

    Args:
        cpt: layer number in mdoel indexing
        model: model containing the layer config
    """
    if len(model.get_config()["layers"][cpt]["inbound_nodes"]) == 0:
        return None
    extra_args = model.get_config()["layers"][cpt]["inbound_nodes"][0][-1]
    if isinstance(extra_args, list):
        return [-5, 5]  # shitty fix for shitty bug
    if "y" in extra_args:
        return {"y": extra_args["y"]}
    elif "clip_value_min" in extra_args:
        return {
            "clip_value_min": extra_args["clip_value_min"],
            "clip_value_max": extra_args["clip_value_max"],
        }
    else:
        return [value for _, value in extra_args.items()]


def retrieve_layer_cpt(model: tf.keras.Model, layer_name: str) -> int:
    """
    retrive the index of the layer in the list model.layers

    Args:
        model: model containing the target layer
        layer_name: target layer name
    """
    for cpt, layer in enumerate(model.layers):
        if layer.name == layer_name:
            return cpt


def call_lambda_layer(
    layer_input: tf.Variable,
    model: tf.keras.Model,
    layer: tf.keras.layers.Layer,
    layer_cpt: int,
) -> tf.Variable:
    """
    this function deals with lambda layers
    the issue is : lambda layers often use parameters
    that are neither weights nor variable and are
    only accessible in the mdoel config
    """
    if not isinstance(layer, tf.keras.layers.Lambda) and not (
        "lambda" in type(layer).__name__.lower()
    ):
        output = layer(layer_input)
        return output
    if layer_cpt == -1:
        layer_cpt = retrieve_layer_cpt(model=model, layer_name=layer.name)
    extra_args = get_lambda_args(cpt=layer_cpt, model=model)
    if extra_args is None:
        return layer(layer_input)
    copied_layer = type(layer).from_config(layer.get_config())
    if "y" in extra_args:
        output = copied_layer(layer_input, extra_args["y"])
    elif "clip_value_min" in extra_args:
        output = copied_layer(
            layer_input, extra_args["clip_value_min"], extra_args["clip_value_max"]
        )
    elif "tf.cast" in layer.name.lower():
        output = layer(
            layer_input,
            dtype=model.get_config()["layers"][layer_cpt]["config"]["dtype"],
        )
    else:
        output = copied_layer(layer_input, *extra_args)
    return output



import tensorflow as tf
from typing import Dict, Tuple


def get_input_tensors_dict(model: tf.keras.Model) -> Dict[str, tf.Tensor]:
    """
    this is a fix for multiple inputs
    """
    output = {}
    cpt = 0
    for layer in model.layers:
        if isinstance(layer, tf.keras.layers.InputLayer):
            output[layer.name] = model.input[cpt]
            cpt += 1
    return output


def get_graph_as_dict(model: tf.keras.Model) -> Tuple[Dict[str, Dict[str, list]], bool]:
    """
    This function returns a dictionnary of the layers and their corresponding
    input layers. This serves the purpose of re-defining the graph with
    new layers.
    """
    istransformer = False
    network_dict = {"input": get_input_tensors_dict(model), "new_output": {}}
    for layer in model.layers:
        for node in layer._outbound_nodes:
            layer_name = node.outbound_layer.name
            if layer_name not in network_dict["input"]:
                network_dict["input"].update({layer_name: [layer.name]})
            else:
                if layer.name not in network_dict["input"][layer_name]:
                    network_dict["input"][layer_name].append(layer.name)
    if network_dict == {"input": {}, "new_output": {}}:
        print("[\033[91mWARNING\033[00m] not supported architecture")
    if isinstance(model.input, (list, tuple)):
        istransformer = True
        for inputs in model.input:
            network_dict["new_output"][inputs.name] = layer.input
    else:
        for layer in model.layers:
            if isinstance(layer, tf.keras.layers.InputLayer):
                network_dict["new_output"].update({layer.name: inputs})
    for layer in model.layers:
        if layer.name not in network_dict["input"]:
            network_dict["input"][layer.name] = []
    return network_dict, istransformer


def reverse_graph(graph: Dict[str, list]) -> Dict[str, list]:
    """
    This function fetches the output layers of each layer of the DNN
    """
    output_dict = {}
    all_keys = []
    for key_1 in list(graph.keys()):
        for key_2 in graph[key_1]:
            if key_2 not in output_dict:
                output_dict.update({key_2: [key_1]})
            else:
                if key_1 not in output_dict[key_2]:
                    output_dict[key_2].append(key_1)
            if key_2 not in all_keys:
                all_keys.append(key_2)
        if key_1 not in all_keys:
            all_keys.append(key_1)
    for key in all_keys:
        if key not in output_dict:
            output_dict[key] = []
    return output_dict
